// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    config::{CpuConfig, CpuConfigType},
    instruction::{MOpRegNum, PRegNum},
    util::range_intersection,
};
use fayalite::{
    memory::{splat_mask, ReadStruct, WriteStruct},
    module::memory_with_loc,
    prelude::*,
};
use std::{mem, ops::Range};

#[hdl(no_static)]
pub struct RenameTableReadPort<C: Type + CpuConfigType> {
    pub addr: MOpRegNum,
    #[hdl(flip)]
    pub data: PRegNum<C>,
}

#[hdl(no_static)]
pub struct RenameTableWritePort<C: Type + CpuConfigType> {
    pub addr: MOpRegNum,
    pub data: PRegNum<C>,
}

#[derive(Clone, Debug)]
pub enum RenameTablePortConfig {
    Read { addr_range: Range<u32> },
    Write { addr_range: Range<u32> },
}

type C = PhantomConst<CpuConfig>;

/// register rename table.
/// all read/write operations are done in the order of `port_configs`.
/// So if `port_configs[0]` is a write and `port_configs[1]` is a read,
/// then the read port will combinatorially return data written by the
/// write port in the *same* clock cycle. However, if `port_configs[0]`
/// is a read and `port_configs[1]` is a write, then the read port will
/// not see the data written by the write port until the *next* clock cycle.
#[hdl_module]
pub fn rename_table(config: PhantomConst<CpuConfig>, port_configs: &[RenameTablePortConfig]) {
    let read_count = port_configs
        .iter()
        .filter(|v| matches!(v, RenameTablePortConfig::Read { .. }))
        .count();
    let write_count = port_configs
        .iter()
        .filter(|v| matches!(v, RenameTablePortConfig::Write { .. }))
        .count();

    #[hdl]
    let cd: ClockDomain = m.input();
    #[hdl]
    let read_ports: Array<RenameTableReadPort<C>> =
        m.input(Array[RenameTableReadPort[config]][read_count]);
    #[hdl]
    let write_ports: Array<RenameTableWritePort<C>> =
        m.input(Array[RenameTableWritePort[config]][write_count]);

    for read_port in read_ports {
        connect(read_port.data, PRegNum[config].const_zero());
    }

    let port_configs_and_indexes = port_configs.iter().scan(
        (0usize, 0),
        |(read_port_index, write_port_index), port_config| {
            Some((
                port_config,
                match port_config {
                    RenameTablePortConfig::Read { .. } => {
                        mem::replace(read_port_index, *read_port_index + 1)
                    }
                    RenameTablePortConfig::Write { .. } => {
                        mem::replace(write_port_index, *write_port_index + 1)
                    }
                },
            ))
        },
    );

    let mut range_transitions = Vec::with_capacity(port_configs.len() * 2);
    for port_config in port_configs {
        let (RenameTablePortConfig::Read { addr_range }
        | RenameTablePortConfig::Write { addr_range }) = port_config;
        range_transitions.push(addr_range.start);
        range_transitions.push(addr_range.end);
    }
    range_transitions.sort_unstable();
    range_transitions.dedup();
    let mut last_range_transition = None;
    for range_transition in range_transitions {
        let Some(last_range_transition) = last_range_transition.replace(range_transition) else {
            continue;
        };
        let cur_addr_range = last_range_transition..range_transition;
        let mut mem = memory_with_loc(
            &if cur_addr_range.len() == 1 {
                format!("mem_{:#x}", cur_addr_range.start)
            } else {
                format!("mem_{:#x}_{:#x}", cur_addr_range.start, cur_addr_range.end)
            },
            PRegNum[config],
            SourceLocation::caller(),
        );
        mem.depth(cur_addr_range.len());
        let addr_in_range = |addr: Expr<MOpRegNum>| {
            if cur_addr_range.len() == 1 {
                addr.value.cmp_eq(cur_addr_range.start)
            } else {
                addr.value.cmp_ge(cur_addr_range.start) & addr.value.cmp_lt(cur_addr_range.end)
            }
        };
        for (port_config, port_index) in port_configs_and_indexes.clone() {
            match port_config {
                RenameTablePortConfig::Read { addr_range } => {
                    if range_intersection(&addr_range, &cur_addr_range).is_none() {
                        continue;
                    }
                    let port = read_ports[port_index];
                    #[hdl]
                    let ReadStruct::<_, _> {
                        addr,
                        en,
                        clk,
                        data,
                    } = mem.new_read_port();
                    connect_any(addr, port.addr.value - cur_addr_range.start);
                    connect(en, addr_in_range(port.addr));
                    connect(clk, cd.clk);
                    #[hdl]
                    if en {
                        connect(port.data, data);
                    }
                }
                RenameTablePortConfig::Write { addr_range } => {
                    if range_intersection(&addr_range, &cur_addr_range).is_none() {
                        continue;
                    }
                    let port = write_ports[port_index];
                    #[hdl]
                    let WriteStruct::<_, _> {
                        addr,
                        en,
                        clk,
                        data,
                        mask,
                    } = mem.new_write_port();
                    connect_any(addr, port.addr.value - cur_addr_range.start);
                    connect(en, addr_in_range(port.addr));
                    connect(clk, cd.clk);
                    connect(data, port.data);
                    connect(mask, splat_mask(Expr::ty(port).data, true.to_expr()));
                }
            }
        }
    }
    for (port_config_index, (port_config, port_index)) in
        port_configs_and_indexes.clone().enumerate()
    {
        let RenameTablePortConfig::Read { addr_range } = port_config else {
            continue;
        };
        let port = read_ports[port_index];
        for (prev_port_config, prev_port_index) in
            port_configs_and_indexes.clone().take(port_config_index)
        {
            let RenameTablePortConfig::Write {
                addr_range: prev_addr_range,
            } = prev_port_config
            else {
                continue;
            };
            if range_intersection(addr_range, prev_addr_range).is_none() {
                continue;
            }
            let prev_port = write_ports[prev_port_index];
            #[hdl]
            if prev_port.addr.cmp_eq(port.addr) {
                connect(port.data, prev_port.data);
            }
        }
    }
}
